import struct
from hks_pynetwork.packet import PacketEncoder, PacketDecoder
from hks_pylib.cryptography.cipherid import CipherID, hash_cls_name
from hks_pylib.cryptography._cipher import HKSCipher

class SecurePacketException(Exception): ...
class CipherTypeMismatch(SecurePacketException): ...


class SecurePacketEncoder(PacketEncoder):
    # A packet generator encrypting by AES-256 and hashing by MD5
    def __init__(self, cipher: HKSCipher):
        self.cipher = cipher

    def __call__(self, payload: bytes):
        payload = self.cipher.encrypt(payload)
        packet = super().pack(payload)

        # SECURE HEADER = TYPE_OF_CIPHER (2 bytes) + NUMBER_OF_PARAMS(1 byte)
        #                 + PARAM1_SIZE + PARAM1 + PARAM2_SIZE + PARAM2 + ...
        # TYPE_OF_CIPHER is the hash value of the cipher class

        secure_header = hash_cls_name(self.cipher) + struct.pack(">B", self.cipher._number_of_params)

        for i in range(self.cipher._number_of_params):
            param = self.cipher.get_param(i)

            if not isinstance(param, bytes):
                raise Exception("Parameter of cipher must be a bytes object")

            param_size = len(param)
            param_struct = "B{}s".format(param_size)
            secure_header += struct.pack(param_struct, param_size, param)

        # Set new header size = old header size + secure header size
        old_header_size = struct.unpack(">H", packet[:2])[0]
        packet = packet[:old_header_size] + secure_header + packet[old_header_size:]
        new_header_size = old_header_size + len(secure_header)
        new_header_size = struct.pack(">H", new_header_size)

        # Return the packet with new header size
        return new_header_size + packet[2:]


class SecurePacketDecoder(PacketDecoder):
    # A packet generator encrypting by AES-256 and hashing by MD5
    def __init__(self, cipher: HKSCipher):
        self.cipher = cipher

    def __call__(self, packet: bytes):
        packet_dict = super().unpack(packet)

        # ORIGINAL HEADER = HEADER_SIZE (2 bytes) + PAYLOAD_SIZE (2 byte)
        original_header_size = 6

        # SECURE HEADER: TYPE_OF_CIPHER (2 bytes) + NUMBER_OF_PARAMS(1 byte)
        #                 + PARAM1_SIZE + PARAM1 + PARAM2_SIZE + PARAM2 + ...
        cipher_hashvalue = packet[original_header_size: original_header_size + 2]
        cipher_type = CipherID.hash2cls(cipher_hashvalue)

        if cipher_type is None:
            raise CipherTypeMismatch("Cipher type is invalid (hash = {})".format(cipher_hashvalue))
        if not isinstance(self.cipher, cipher_type):
            raise CipherTypeMismatch("Cipher type mismatches (expected {}, but received {})".format(
                type(self.cipher).__name__,
                cipher_type.__name__
            ))

        number_of_params = struct.unpack(">B", packet[original_header_size + 2: original_header_size + 3])[0]
        current_index = original_header_size + 3
        for i in range(number_of_params):
            param_size = struct.unpack(">B", packet[current_index: current_index + 1])[0]
            current_index += 1

            param = struct.unpack(">{}s".format(param_size), packet[current_index: current_index + param_size])[0]
            current_index += param_size

            self.cipher.set_param(i, param)
        self.cipher.reset(False)
        packet_dict["payload"] = self.cipher.decrypt(packet_dict["payload"])
        return packet_dict



import copy
import time
import errno
import socket
import threading

from hks_pylib.logger import LoggerGenerator
from hks_pylib.cryptography._cipher import HKSCipher
from hks_pylib.cryptography.cipherid import CipherID

from hks_pynetwork.packet_buffer import PacketBuffer
from hks_pynetwork.secure_packet import SecurePacketEncoder, SecurePacketDecoder

# Exception
from hks_pylib.cryptography.symmetrics import DecryptFailed
from hks_pynetwork.packet import CannotExtractPacket
from hks_pynetwork.secure_packet import CipherTypeMismatch


class STCPSocketException(Exception): ...
class STCPSocketClosed(STCPSocketException): ...
class STCPSocketTimeout(STCPSocketException): ...


class STCPSocket(object):
    DEFAULT_TIME_OUT = 0.1

    def __init__(
                    self,
                    cipher: HKSCipher,
                    buffer_size: int = 1024,
                    reload_time: float = 0.001,
                    logger_generator: LoggerGenerator = ...,
                    display: dict = ...,
                    name: str="STCP Socket Listener",
                ):
        assert isinstance(cipher, HKSCipher)
        assert isinstance(buffer_size, int)
        assert isinstance(reload_time, float)
        assert isinstance(logger_generator, LoggerGenerator)
        assert isinstance(display, dict)
        assert isinstance(name, str)

        self._logger_generator = logger_generator
        self._display = display
        self.__print = self._logger_generator.generate(name, self._display)

        self.__print("dev", "debug", "Initialized with cipher {}".format(CipherID.cls2name(type(cipher))))
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__cipher = cipher
        self.__cipher.reset()
        self.set_reload_time(reload_time)
        self.__recv_timeout = None

        self.__packet_encoder = SecurePacketEncoder(self.__cipher)
        self.__packet_decoder = SecurePacketDecoder(self.__cipher)
        self.__buffer = None
        self.__buffer_size = buffer_size

        self.__buffer_available = threading.Event()
        self._stop_auto_recv = False
        self._prepare_close = False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return self._socket.__exit__(args)

    def _start_auto_recv(self):
        self.__print("dev", "info", "Start automatically receiving...")
        while True:
            try:
                data = self._socket.recv(self.__buffer_size)
            except socket.error as e:
                if isinstance(e, socket.timeout) and not self._stop_auto_recv:
                    continue

                # self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()
                if e.errno in (errno.ECONNRESET, errno.ECONNABORTED, errno.ECONNREFUSED):
                    self.__print("dev", "info", "Automatic receiving closes normally ({})".format(repr(e)))
                    break
                elif isinstance(e, socket.timeout):
                    self.__print("dev", "info", "Automatic receiving closes normally (timeout and socket closed)")
                    break
                elif isinstance(e, OSError) and e.errno == 10038:
                    self.__print("dev", "info", "Automatic receiving closes normally (remote socket closed)")
                    break
                else:
                    self.__print("dev", "error", "Connection stops suddenly ({})".format(repr(e)))
                    raise e
            except Exception as e:
                # self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()
                self.__print("dev", "error", "Unknown error in automatically receiving ({})".format(repr(e)))
                raise e
            else:
                if not data:  # when be closed, socket will receive infinite empty packet
                    self._socket.close()
                    self.__print("dev", "info", "Automatic receiving closes normally (remote socket closed)")
                    break
                self.__buffer.push(data)
                self.__buffer_available.set()
        self.__buffer_available.set()
        self.__print("dev", "info", "Stop automatic receiving...")

    def set_reload_time(self, value: float):
        self.__print("dev", "debug", "Set reload time to {}".format(value))
        self.__reload_time = value

    def set_recv_timeout(self, value: float):
        self.__print("dev", "debug", "Set receive timeout to {}".format(value))
        self.__recv_timeout = value

    def settimeout(self, value: float):
        self.__print("dev", "debug", "Set timeout to {}".format(value))
        return self._socket.settimeout(value)

    def setreloadtime(self, reloadtime: float):
        self.__reload_time = reloadtime

    def recv(self) -> bytes:
        data = b''
        if self.__recv_timeout is not None:
            start_recv_time = time.time()

        while not data:
            if self.isclosed() and len(self.__buffer) == 0:
                raise STCPSocketClosed("Connection closed")

            try:
                data = self.__buffer.pop()
            except CannotExtractPacket as e:
                # Not enough length of packet
                continue
            except CipherTypeMismatch as e:
                # Cipher's type mismatch
                self.__print("user", "warning", "Detect an abnormal packet")
                self.__print("dev", "warning", "Detect an abnormal packet ({})".format(repr(e)))
                break
            except DecryptFailed as e:
                # Cannot decrypt the packet: authentication failed, wrong parameters, ...
                self.__print("user", "warning", "Detect an abnormal packet")
                self.__print("dev", "warning", "Detect an abnormal packet ({})".format(repr(e)))
                break
            except Exception as e:
                self.__print("user", "info", "Unknown error in recv() method")
                self.__print("dev", "error", "Unknown error in recv() method ({})".format(repr(e)))
                break
            finally:
                if not data:
                    if self.__recv_timeout is not None and time.time() - start_recv_time > self.__recv_timeout:
                        if self.__recv_timeout != 0:
                            raise STCPSocketTimeout("Receiving exceeds timeout")
                        else:
                            raise STCPSocketTimeout("Method recv() can not return the value immediately")
                    time.sleep(self.__reload_time)

                    if self.__recv_timeout is None and len(self.__buffer) == 0 and not self.isclosed():
                        self.__buffer_available.wait()

        self.__buffer_available.clear()
        return data

    def send(self, data) -> int:
        self.__cipher.reset()
        packet = self.__packet_encoder(data)
        return self._socket.send(packet)

    def sendall(self, data) -> int:
        self.__cipher.reset()
        packet = self.__packet_encoder(data)
        return self._socket.sendall(packet)

    def bind(self, address):
        return self._socket.bind(address)

    def listen(self, __backlog: int = 0):
        self.__print("user", "info", "Server listen...")
        self.__print("dev", "info", "Server listen...")
        return self._socket.listen(__backlog)

    def accept(self):
        socket, addr = self._socket.accept()
        socket.settimeout(STCPSocket.DEFAULT_TIME_OUT)  # timeout for non-blocking socket
        self.__print("user", "info", "Server accept {}".format(addr))
        self.__print("dev", "info", "Server accept {}".format(addr))
        socket = self._fromsocket(socket, addr, start_serve=True)
        return socket, addr

    def connect(self, address):
        ret = self._socket.connect(address)
        self.__print("user", "info", "Connect to server {} successfully".format(address))
        self.__print("dev", "info", "Connect to server {} successfully".format(address))
        self.__print("dev", "debug", "Transform to STCP Socket {}".format(address))

        self.__print = self._logger_generator.generate(f"STCP Socket {address}", self._display)
        self.settimeout(STCPSocket.DEFAULT_TIME_OUT)  # timeout for non-blocking socket
        
        self._stop_auto_recv = False
        server = threading.Thread(target=self._start_auto_recv)
        server.start()

        self.__buffer = PacketBuffer(
            decoder=self.__packet_decoder,
            identifier=address,
            logger_generator=self._logger_generator,
            display=self._display
        )
        return ret

    def close(self):
        if self.__buffer is None:  # STCP Listener
            self._socket.close()
        else:
            self.__buffer_available.set()
            self._stop_auto_recv = True
            self._socket.close()
        self.__print("dev", "info", "Closed")

    def isclosed(self):
        return self._socket._closed

    def _fromsocket(self, socket: socket.socket, address, start_serve=True):
        cipher = copy.copy(self.__cipher)
        new_socket = STCPSocket(
            cipher=cipher, 
            buffer_size=self.__buffer_size,
            logger_generator=self._logger_generator,
            display=self._display,
            name=f"STCP Socket {address}"
        )
        new_socket._socket = socket
        new_socket.__buffer = PacketBuffer(
            decoder=new_socket.__packet_decoder,
            identifier=address,
            logger_generator=self._logger_generator,
            display=self._display
        )
        if start_serve:
            new_socket._stop_auto_recv = False
            server = threading.Thread(target=new_socket._start_auto_recv)
            server.start()
        return new_socket



import random
import threading
from hks_pylib.logger import LoggerGenerator
from hks_pynetwork.external import STCPSocket, STCPSocketClosed

class ChannelException(Exception): ...
class ChannelSlotError(ChannelException): ...
class ChannelFullSlots(ChannelSlotError): ...
class ChannelObjectExists(ChannelException): ...
class ChannelMessageFormatError(ChannelException): ...
class ChannelClosed(ChannelException): ...


class ChannelBuffer(object):
    def __init__(self):
        self._buffer = []
        self.__lock = threading.Lock()

    def push(self, source: str, message: bytes, obj: object = None):
        self.__lock.acquire()
        self._buffer.append({"source": source, "message": message, "obj": obj})
        self.__lock.release()

    def pop(self, source: str = None):
        self.__lock.acquire()
        if len(self._buffer) == 0:
            return None, None, None

        idx = 0
        if source is not None:
            for i, packet in enumerate(self._buffer):
                if packet["source"] == source:
                    idx = i
                    break

        packet = self._buffer[idx]
        del self._buffer[idx]

        self.__lock.release()
        return packet["source"], packet["message"], packet["obj"]

    def __len__(self):
        return len(self._buffer)


class LocalNode(object):
    nodes = []
    node_names = []
    MAX_NODES = 8
    lock = threading.Lock()

    def __init__(self,
            name: str = None,
            logger_generator: LoggerGenerator = ...,
            display: dict = ...
        ):
        if name is None:
            while name is None or name in LocalNode.node_names:
                name = str(random.randint(1000000, 9999999))

        if name in LocalNode.node_names:
            raise ChannelObjectExists(f"Name {name} is in use")

        if len(LocalNode.nodes) >= LocalNode.MAX_NODES:
            raise ChannelFullSlots("No available slot in Local Node list")

        LocalNode.lock.acquire()
        LocalNode.node_names.append(name)
        LocalNode.nodes.append(self)
        LocalNode.lock.release()
        self.name = name
        self._buffer = ChannelBuffer()
        self._closed = False
        self._buffer_available = threading.Event()
        self.__send_lock = threading.Lock()
        self.__recv_lock = threading.Lock()
        self._print = logger_generator.generate(
            name,
            display
        )
        self._print("dev", "info", "The node join to Local Node")

    def send(self, destination_name: str, message: bytes, obj: object = None):
        self.__send_lock.acquire()
        if isinstance(message, bytes) is False:
            raise Exception("Message must be a bytes object")

        if self._closed:
            raise ChannelClosed("Channel closed")

        try:
            slot_of_destination = LocalNode.node_names.index(destination_name)
            destination_node: LocalNode = LocalNode.nodes[slot_of_destination]
        except ValueError:
            raise ChannelSlotError(f"Channel name {destination_name} doesn't exist")

        destination_node._buffer.push(self.name, message, obj)
        destination_node._buffer_available.set()
        self.__send_lock.release()

    def recv(self, source=None):
        if self._closed:
            raise ChannelClosed("Channel closed")

        if len(self._buffer) == 0 and not self._closed:
            self._buffer_available.wait()
        self._buffer_available.clear()

        if self._closed:
            raise ChannelClosed("Channel closed")

        source, msg, obj = self._buffer.pop(source)
        return source, msg, obj

    def close(self):
        if not self._closed:
            self.__recv_lock.acquire()
            self._closed = True
            self._buffer_available.set()
            my_slot = LocalNode.node_names.index(self.name)
            del LocalNode.node_names[my_slot]
            del LocalNode.nodes[my_slot]
            del self._buffer
            self.name = None
            self.__recv_lock.release()
            self._print("dev", "info", "The node leaves from Local Node")
        else:
            self._print("dev", "warning", "Call close() when channel closed")


class ForwardNode(LocalNode):
    def __init__(
                    self,
                    node: LocalNode = None,
                    socket: STCPSocket = None,
                    name: str = None,
                    implicated_die: bool = False,
                    logger_generator: LoggerGenerator = ...,
                    display: tuple = ...
                ):
        assert node is None or isinstance(node, LocalNode)
        assert socket is None or isinstance(socket, STCPSocket)
            
        self._node = node
        self._socket = socket
        super().__init__(name, logger_generator, display)

        self._implicated_die = implicated_die
        if display is None:
            display = socket.__print.display

        self.__print = logger_generator.generate(f"{self.name}", display)
        self._one_thread_stop = threading.Event()

    def set_node(self, node):
        assert node is None or isinstance(node, LocalNode)
        self._node = node
    
    def set_socket(self, socket):
        assert socket is None or isinstance(socket, STCPSocket)
        self._socket = socket

    def start(self):
        self.__print("dev", "info", "Start forward between local node and remote node")
        if not self._node or not self._socket:
            raise Exception("ForwardNode must be initilized with both node and socket")

        t1 = threading.Thread(target=self._wait_message_from_node)
        t1.setDaemon(True)
        t1.start()

        t2 = threading.Thread(target=self._wait_message_from_remote)
        t2.setDaemon(True)
        t2.start()

        self._one_thread_stop.wait()
        self.close()
        if self._implicated_die:
            self._node.close()
            self._socket.close()
        else:
            if self._node._buffer_available.is_set() is False:
                self._node._buffer_available.set()

        self.__print("dev", "info", "Stop completely")

    def _wait_message_from_remote(self):
        while not self._closed:
            try:
                data = self._socket.recv()
            except STCPSocketClosed:
                self.__print("dev", "info", "Waiting message from remote close because STCP Socket closed")
                break
            except Exception as e:
                self.__print("user", "info", "Unknown error in waiting message from remote client")
                self.__print("dev", "error", 
                    "Unknown error in waiting message from remote client ({})".format(repr(e))
                )
                break
            if data:
                try:
                    super().send(self._node.name, data)
                except ChannelClosed:
                    self.__print("dev", "info", "Waiting message from remote close because channel closed")
                    break
                except Exception as e:
                    self.__print("user", "info", "Unknown error in sending data to node")
                    self.__print("dev", "error", 
                        "Unknown error in sending data to node ({})".format(repr(e))
                    )
                    break
        self._one_thread_stop.set()
        self.__print("dev", "info", "Waiting from remote ended")

    def _wait_message_from_node(self):
        while not self._closed:
            try:
                _, message, _ = super().recv()
            except AttributeError as e:  # after close forwarder, it dont have buffer attribute --> error
                self.__print("dev", "warning", "{} in waiting message from local node".format(e))
                self.__print("user", "info", "Waiting message from local node close because channel closed")
                break
            except ChannelClosed:
                self.__print("user", "info", "Waiting message from local node close because channel closed")
                self.__print("dev", "info", "Waiting message from local node close because channel closed")
                break
            except Exception as e:
                self.__print("user", "info", "Unknown error in waiting message from local node")
                self.__print("dev", "error", 
                    "Unknown error in waiting message from local node ({})".format(repr(e))
                )
                break
            if message:
                try:
                    self._socket.send(message)
                except STCPSocketClosed:
                    self.__print("dev", "info", " Waiting message from local node close because STCP Socket closed")
                    break

        self._one_thread_stop.set()
        self.__print("dev", "info", "Waiting from node ended")

    def send(self, received_node_name, message):
        raise NotImplementedError

    def recv(self):
        raise NotImplementedError



import struct
import threading

from hks_pylib.logger import LoggerGenerator
from hks_pynetwork.secure_packet import CipherTypeMismatch
from hks_pynetwork.packet import CannotExtractPacket, PacketDecoder

class PacketBufferException(Exception): ...
class PacketBufferOverflow(PacketBufferException): ...


class PacketBuffer():
    def __init__(
                    self,
                    decoder: PacketDecoder,
                    identifier=None,
                    logger_generator: LoggerGenerator = ...,
                    display: dict = ...
                ) -> None:
        self._buffer = []
        self._packet_decoder = decoder
        prefix = "PacketBuffer"
        if identifier is not None:
            prefix = f"PacketBuffer {identifier}"
        self.__print = logger_generator.generate(prefix, display)
        self._current_packet = b""
        self._current_packet_size = 0
        self._expected_current_packet_size = 0
        self._push_lock = threading.Lock()

    def push(self, packet: bytes, append_to_end=True):
        self._push_lock.acquire()
        if append_to_end:
            self._buffer.append(packet)
        else:
            self._buffer.insert(0, packet)
        self._push_lock.release()

    def pop(self):
        try:
            if len(self._buffer) == 0:
                return b""

            self._current_packet_size += len(self._buffer[0])
            self._current_packet += self._buffer[0]
            del self._buffer[0]

            if self._packet_decoder is None:
                ret = self._current_packet
                self._current_packet = b""
                self._current_packet_size = 0
                self._expected_current_packet_size = 0
                return ret

            if self._expected_current_packet_size == 0:
                try:
                    packet_dict = self._packet_decoder._decode_header(self._current_packet)
                    self._expected_current_packet_size = packet_dict["payload_size"] + packet_dict["header_size"]
                except CannotExtractPacket:
                    return b""
                except struct.error:
                    return b""
                except Exception as e:
                    self.__print("dev", "warning", "Something error when decode a packet ({})".format(repr(e)))
                    return b""

            if self._current_packet_size < self._expected_current_packet_size:
                return b""

            # if self._current_packet == b"":
            #     self._current_packet_size = 0
            #     self._expected_current_packet_size = 0
            #     return b""

            try:
                packet_dict = self._packet_decoder(self._current_packet)
            except CipherTypeMismatch as e:
                raise e
            finally:
                if self._expected_current_packet_size < self._current_packet_size:
                    apart_of_next_packet = self._current_packet[self._expected_current_packet_size:]
                    self.push(apart_of_next_packet, append_to_end=False)

                self._current_packet = b""
                self._current_packet_size = 0
                self._expected_current_packet_size = 0
            return packet_dict["payload"]
        except Exception as e:
            # self.__print("dev", "error", "An error occurs when calling pop() ({})".format(repr(e)))
            raise e

    def __len__(self):
        return len(self._buffer)



import struct

MAX_LENGTH = 65535 + 4294967295


class PacketException(Exception): ...
class CannotExtractPacket(PacketException): ...


class PacketEncoder(object):
    @staticmethod
    def pack(payload: bytes):
        if isinstance(payload, bytes) is False:
            raise Exception("Payload must be a byte object")

        # PACKET = HEADER + PAYLOAD
        # HEADER = HEADER_SIZE(2 bytes) + PAYLOAD_SIZE(4 byte) + OPTIONAL_HEADER
        # ==> PACKET = HERDER_SIZE + PAYLOAD_SIZE + OPTIONAL_HEADER + PAYLOAD

        # header struct except for header_size
        header_struct = ">HI"
        header_dummy = struct.pack(header_struct, 0, 0)
        header_size = len(header_dummy)
        header = struct.pack(header_struct, header_size, len(payload))

        return header + payload


class PacketDecoder(object):
    @staticmethod
    def _decode_header(packet: bytes):
        if isinstance(packet, bytes) is False:
            raise Exception("Packet must be a bytes object")

        header_size, payload_size = struct.unpack(">HI", packet[:6])
        if len(packet) < header_size:
            raise CannotExtractPacket("Incomplete header")

        header_dict = {}
        header_dict["header_size"] = header_size
        header_dict["payload_size"] = payload_size

        return header_dict

    @staticmethod
    def unpack(packet: bytes):
        if isinstance(packet, bytes) is False:
            raise Exception("Packet must be a bytes object")

        try:
            header = PacketDecoder._decode_header(packet)
        except Exception as e:
            raise e

        header_size, payload_size = header["header_size"], header["payload_size"]
        if len(packet) < header_size + payload_size:
            raise CannotExtractPacket("Incomplete packet")

        packet_dict = header
        packet_dict["payload"] = packet[header_size: header_size + payload_size]
        packet_dict["packet_size"] = header_size + payload_size

        return packet_dict
